//Name        : DensityBounder.
//
//Author      : Rahil Baber.
//
//Date        : 21st April 2010.
//
//Description : Determines an upper bound for Turan densities of 3-uniform hypergraphs. It
//              creates a SDP problem to determine the bound. The SDP problem can then be
//              solved by existing applications. In order to turn the resulting floating
//              point output of the SDP solver into a proof, the algorithm approximates it
//              into a rational solution which allows the calculation to be made without
//              introducing rounding errors.
//
//Notes       : We assume the type double is of size 8 bytes and is equivalent to binary64 as
//              defined in the IEEE 754-2008 standard. Because the solution to the SDP
//              problem will vary depending on the solver being used and the system it is run
//              on, the results may vary slightly from those given. We provide online the
//              ".soln" files we used to achieve the bounds to allow others to verify our
//              claims precisely and consistently, without any dependence on an external SDP
//              solver or floating point arithmetic. To do so simply run the program on the
//              provided ".soln" file and choose option 3 when prompted.

#include <iostream>
#include <iomanip>
#include <fstream>
#include <stdlib.h>
#include <math.h>
using namespace std;

//We use #define to allow users to quickly change parameters to solve different problems.

#define _HYPERGRAPHS_DO_JUMP_
//#define _FORBIDDING_K4MINUS_
//#define _RAZBOROV_FORBIDDING_K4_

//The size of a large integer in bytes. Used by the CLargeInt class, and functions which
//begin with the prefix "LI_". LI_MAX should be greater than or equal to sizeof(long) to
//allow "long"s to be converted into "CLargeInt"s. (sizeof(long) is usually 4 or 8 on 32 and
//64  bit systems respectively.) To conserve memory LI_MAX should be as small as possible.
const long LI_MAX = 10;

//Used to approximate a double into a fraction with denominator 2^accuracy. The numerator is
//stored as a CLargeInt and so the upper limit of "accuracy" depends on the size of a
//CLargeInt. The lower limit of "accuracy" is 0. The default value given below is used by
//RationalizeSolution(). The value is overwritten by ReadSolnFile() with the value stored in
//the ".soln" file.
long accuracy = 20;

//Used by RationalizeSolution() to avoid division by zero, by classifying any double with a
//magnitude at most "tolerance" as zero.
double tolerance = 0.0001;

//Used by DisplayDensity() to calculate the Turan density to "noOfDecimalPlaces" decimal
//places.
long noOfDecimalPlaces = 6;

//Set the data needed to calculate the upper bound for the Turan density for the
//"Hypergraphs Do Jump" problem. The bound should be 0.2299.
#ifdef _HYPERGRAPHS_DO_JUMP_

//Set the name of the file the SDP problem will be written to. The problem will be written in
//the ".dat-s" format.
char filenameSDP[] = "HypergraphsDoJump.dat-s";

//Set the name of the file where the solution to the SDP problem resides. The program expects
//the file to be a text file containing the value of the variables in order separated by
//spaces. The ".ini-s" file created by the stand-alone CSDP solver is sufficient (the data
//after the variable values is ignored).
char filenameOutput[]  = "HypergraphsDoJump.out";

//Set the name of the file where the data, to construct a rationalized solution to the SDP
//problem, is stored. The file is a text file containing large integers separated by spaces.
//The first entry is the value of the "accuracy" variable, the remaining integers are entries
//to matrices. If M is such a matrix then we construct the positive semidefinite matrix we
//need to calculate the upper bound of the Turan density by taking $(M^T M) / 4^{accuracy}$.
//Note that $M^T M$ is positive semidefinite for all matrices M.
char filenameSoln[] = "HypergraphsDoJump.soln";

//The order of the graphs $H_i$. Denoted by $l$ in the thesis. Due to the way the function
//Generate() has been implemented orderOfH can be at most 9. In order for the density
//calculation to be valid orderOfH must be at least 3. To ensure overflow doesn't occur
//2*(maxNoOfFlags^2)*(orderOfH!) should not exceed the largest integer storable in a "long".
const long orderOfH = 7;

//The size of our forbidden family $\mathcal{F}$.
const long noOfForbiddenGraphs = 5;

//The number of types we will be using to construct flags.
const long noOfTypes = 6;

//An upper bound for $|\mathcal{H}|$ the number of admissible graphs of order "orderOfH".
//The value is hard coded in to simplify the code by avoiding dynamic memory allocation. To
//conserve memory this should be as close to $|\mathcal{H}|$ as possible.
const long maxNoOfGraphs = 5000;

//An upper bound for the number of flags we will generate by any of the "noOfTypes" types.
//The value is hard coded in to simplify the code by avoiding dynamic memory allocation. To
//conserve memory it should be as small as possible.
const long maxNoOfFlags = 150;

//The edge set of the graphs in $\mathcal{F}$. Uninitialized elements of the array are set to
//zero (by default), which will be used to indicate the end of the edge set.
long ForbiddenEdgeSet[noOfForbiddenGraphs][14] = {
    {123, 124, 134},
    {123, 124, 125, 345},
    {123, 124, 235, 145, 345},
    {123, 135, 145, 245, 126, 246, 346, 356, 237, 147, 347, 257, 167},
    {123, 124, 135, 145, 236, 346, 256, 456, 247, 347, 257, 357, 167}};

//The order of the graphs in $\mathcal{F}$.
long ForbiddenOrder[noOfForbiddenGraphs] = {4, 5, 5, 7, 7};

//If bForbidOnlyInducedSubgraphs is set to false, the admissible graphs are those which do
//not contain a subgraph from our forbidden family $\mathcal{F}$. If set to true then the
//admissible graphs are those which do not contain an *induced* subgraph from $\mathcal{F}$. 
bool bForbidOnlyInducedSubgraphs = false;

//The edge set of the types we will be considering. Uninitialized elements of the array are
//set to zero (by default), which will be used to indicate the end of the edge set.
long TypeEdgeSet[noOfTypes][5] = {
    {},
    {},
    {123},
    {123, 124, 135},
    {123, 124, 345},
    {123, 124, 135, 245}};

//The order of the types.
long TypeOrder[noOfTypes] = {1,3,3,5,5,5};

//The order of the flags to be generated by the respective types. The values should satisfy
//FlagOrder[n] <= (orderOfH+TypeOrder[n])/2.
long FlagOrder[noOfTypes] = {4,5,5,6,6,6};

#endif

//Similarly we set the data needed to calculate upper bound for the Turan density when we
//forbid graphs to have $K_4^-$ as subgraph. The bound should be 0.2871.
#ifdef _FORBIDDING_K4MINUS_

char       filenameSDP[]       = "K4-.dat-s";
char       filenameOutput[]    = "K4-.out";
char       filenameSoln[]      = "K4-.soln";
const long orderOfH            = 7;
const long noOfForbiddenGraphs = 1;
const long noOfTypes           = 4;
const long maxNoOfGraphs       = 8200;
const long maxNoOfFlags        = 150;

long ForbiddenEdgeSet[noOfForbiddenGraphs][4] = {
    {123,124,134}};
long ForbiddenOrder[noOfForbiddenGraphs] = {4};
bool bForbidOnlyInducedSubgraphs = false;
long TypeEdgeSet[noOfTypes][4] = {
    {},
    {123},
    {123},
    {123, 124, 125}};
long TypeOrder[noOfTypes] = {3,3,4,5};
long FlagOrder[noOfTypes] = {5,5,5,6};

#endif

//Set the data needed to calculate an upper bound for the Turan density when we forbid $K_4$.
//The bound of 0.561666 was given by Razborov in the paper "On 3-hypergraphs with forbidden
//4-vertex configurations".
#ifdef _RAZBOROV_FORBIDDING_K4_

char       filenameSDP[]       = "RazborovK4.dat-s";
char       filenameOutput[]    = "RazborovK4.out";
char       filenameSoln[]      = "RazborovK4.soln";
const long orderOfH            = 6;
const long noOfForbiddenGraphs = 1;
const long noOfTypes           = 6;
const long maxNoOfGraphs       = 1000;
const long maxNoOfFlags        = 100;

long ForbiddenEdgeSet[noOfForbiddenGraphs][5] = {
    {123, 124, 134, 234}};
long ForbiddenOrder[noOfForbiddenGraphs] = {4};
bool bForbidOnlyInducedSubgraphs = true;
long TypeEdgeSet[noOfTypes][4] = {
    {},
    {},
    {},
    {123},
    {123, 124},
    {123, 124, 134}};
long TypeOrder[noOfTypes] = {0,2,4,4,4,4};
long FlagOrder[noOfTypes] = {3,4,5,5,5,5};

#endif

//Factorial[n] = n!. Initialized by a call to Init().
long Factorial[orderOfH+1];

//Iso[][] is an array (orderOfH! by orderOfH) of all permutations of {0,1,...,orderOfH-1}.
//Iso[p][n] is the image of n under permutation p. Iso[][] is ordered in such a way that
//{Iso[p][] : p<=r!} is the set of permutations in which only {0,1,..,r-1} are permuted and
//{r,...,orderOfH-1} are left in place. Iso[][] is initialized by Init().
long** Iso = NULL;

//CLargeInt is a class that represents large integers. It is used to approximate the floating
//point output of the semidefinite program into rationals. We then can check the solution
//satisfies the constraints without introducing any rounding errors. Initialization to a
//small integer can be done using the function LI_Assign(). Basic arithmetic operations can
//be done using LI_Add(), LI_Subtract(), LI_Multiply(), LI_Divide(). Comparison is done via
//LI_Compare(). To display the decimal representation of the integer use the "<<" operator.
//The function ReadLargeInt() reads large integers from an input stream.
class CLargeInt
{
public:
    //Data[] holds the binary representation of the magnitude of the integer (in
    //little-endian order). Specifically the magnitude of the integer =
    //Data[LI_MAX-1]*256^(LI_MAX-1) + ... + Data[2]*256^2 + Data[1]*256 + Data[0].
    unsigned char Data[LI_MAX];

    //bPositive holds the sign of the integer. If the integer is 0, bPositive can be true or
    //false. If the integer is not 0 then
    //bPositive = true  if the integer is > 0,
    //            false if it is          < 0.
    bool bPositive;

    //Displays the integer using the "<<" operator.
    friend ostream& operator<<(ostream& out, const CLargeInt& L);
};

//An array of size maxNoOfGraphs. Memory for this array is allocated in Init() and
//deallocated in Cleanup(). Used in CalculateDensity() to determine the inequality
//constraints on the upper bound of the Turan density. It is also referenced by 
//DisplayDensity().
CLargeInt* Ineq = NULL;

//CGraph is a class representing a 3-uniform hypergraph (with at most orderOfH number of
//vertices). Because C++ indexes arrays starting from 0, we take our vertex set to be
//{0,1,...,nV-1}, where nV is the number of vertices in our graph. We can represent a flag as
//(g, n) a CGraph g and an integer n, by taking the labelled vertices to be 0,1,...,n-1 and
//to avoid confusion we label them 0,1,...,n-1 respectively (instead of 1,2,...,n).
class CGraph
{
public:
    //Number of vertices.
    long nV;

    //The adjacency matrix.
    //M[i][j][k] = true  if edge ijk is in the graph,
    //             false if edge ijk is not in the graph.
    bool M[orderOfH][orderOfH][orderOfH];

    //Displays the graph using the "<<" operator.
    friend ostream& operator<<(ostream& out, const CGraph& g);
};

//The forbidden family $\mathcal{F}$. Initialized in main() using ForbiddenEdgeSet[][] and
//ForbiddenOrder[].
CGraph ForbiddenGraph[noOfForbiddenGraphs];

//H[] is an array representing the family $\mathcal{H}$ of admissible graphs of order
//orderOfH. Memory for this array is allocated in Init() and deallocated in Cleanup(). It is
//allocated to be of size maxNoOfGraphs though not all of the array may be used as 
//maxNoOfGraphs is just an upper bound for $|\mathcal{H}|$. H[] is initialized by a call to
//GenerateH().
CGraph* H = NULL;

//The number of graphs in H[] found by GenerateH().
long noOfH;

//TempList1[] and TempList2[] are arrays used by GenerateH() and GenerateFlags(). Memory for
//these arrays is allocated in Init() and deallocated in Cleanup(). They are allocated to be
//of size maxNoOfGraphs.
CGraph* TempList1 = NULL;
CGraph* TempList2 = NULL;

//A structure that stores information about flags generated from a type.
class CFlagInfo
{
public:
    //The graph of the type that the labelled vertices of the flags induce. Formally the type
    //is (Type, Type.nV).
    CGraph Type;

    //FlagList[] defines a list of flags (FlagList[], Type.nV) for which the labelled
    //vertices induce the graph Type, i.e. FlagList[n].M[i][j][k] should be identical to
    //Type.M[i][j][k] for all i,j,k < Type.nV. 
    CGraph FlagList[maxNoOfFlags];
    
    //The number of flags in FlagList[].
    long noOfFlags;

    //BasisSize[], BasisElementSize[][], BasisElementFlags[][][], store information about a
    //new basis of flags we create that has a property which reduces the number of variables
    //in the SDP problem.

    //BasisSize[0] is the dimension of the elements that are invariant under the
    //automorphism group of Type. BasisSize[1] is the dimension of the elements that are not
    //invariant.
    long BasisSize[2];

    //An element of the basis is a linear combination of flags (with coefficients 1 or -1).
    //How many flags the element is composed of is stored in BasisElementSize[r][i] (where 
    //r=0 if the element is invariant, r=1 if it isn't, and 0 <= i < BasisSize[r] indexes
    //elements in the basis). The indices of where the flags, from the element, appear in
    //FlagList[] is stored in BasisElementFlags[r][i][]. The coefficients of the flags can
    //be determined based on whether we are looking at a invariant element or not. If the
    //element is invariant we take all the coefficients to be 1. If it is not invariant we
    //take the coefficient of the first flag to be 1 and the coefficient of the second to be
    //-1.
    long BasisElementSize[2][maxNoOfFlags];
    long BasisElementFlags[2][maxNoOfFlags][maxNoOfFlags];
};

//FInfo is an array of size noOfTypes. Memory is allocated in Init() and deallocated in
//Cleanup(). FInfo[].Type is initialized in main() using TypeEdgeSet[][] and TypeOrder[].
//FInfo[].FlagList[] is initialized using GenerateFlags() and FlagOrder[].
//FInfo[].BasisSize[], FInfo[].BasisElementSize[], FInfo[].BasisElementFlags[] are
//initialized by GenerateBasis(). The product of two flags is calculated by
//CalculateFlagProducts(), and the results stored in the CFlagProducts object pointed to by
//pFProd.
CFlagInfo* FInfo = NULL;

//A structure that stores the result of multiplying two flags from a given CFlagInfo
//structure. Each CFlagProducts object takes up a large amount of memory, so is defined as a
//separate object, and only one such object will be created.
class CFlagProducts
{
public:
    //$\mathbf{E}_{\theta\in\Theta_i}(p(FlagList[a], FlagList[b], \theta ; H[i]))$ is a value
    //which can be represented as an integer over (orderOfH!). We store this integer in
    //Data[a][b][i]. This allows us to calculate
    //$(orderOfH!)[[ FlagList[a]FlagList[b] ]]_{Type}$ as 
    //$\sum_{i=0}^{noOfFlags-1} Data[a][b][i] p(H[i] ; G)$.
    long Data[maxNoOfFlags][maxNoOfFlags][maxNoOfGraphs];
};

//pFProd is a pointer to a single CFlagProducts object. The object is allocated in Init() and
//deallocated in Cleanup(). It is initialized from data held in a CFlagInfo object by the
//function CalculateFlagProducts().
CFlagProducts* pFProd = NULL;

//A structure to hold the floating point output of the semidefinite program and its rational
//approximation.
class CSolution
{
public:
    //The dimension of the matrices.
    long dim;

    //Holds the output to the semidefinite program.
    double Matrix[maxNoOfFlags][maxNoOfFlags];

    //The rational approximation of "Matrix" is given by (LMatrix^T LMatrix)/(4^accuracy)
    //which must be positive semidefinite.
    CLargeInt LMatrix[maxNoOfFlags][maxNoOfFlags];
};

//An array of size 2*noOfTypes, used to hold the solution to the SDP problem. Memory is
//allocated in Init() and deallocated in Cleanup(). Soln[].dim is given by the
//FInfo[].BaseSize[]. Soln[].Matrix[][] is initialized in CreateSolnFile() by reading the
//file containing the output from the SDP solver. Soln[].LMatrix[][] is either initialized
//from Soln[].Matrix[][] by RationalizeSolution(), or is initialized directly from data held
//in a ".soln" file via ReadSolnFile().
CSolution* Soln = NULL;

//Deallocates any memory that Init() allocated.
void Cleanup()
{
    long i;

    //Free Iso[][].
    if(Iso!=NULL)
    {
        for(i=0;i<Factorial[orderOfH];i++)
            if(Iso[i]!=NULL)
            {
                free(Iso[i]);
                Iso[i] = NULL;
            }

        free(Iso);
        Iso = NULL;
    }

    //Free Ineq[].
    if(Ineq!=NULL)
    {
        free(Ineq);
        Ineq = NULL;
    }

    //Free H[].
    if(H!=NULL)
    {
        free(H);
        H = NULL;
    }

    //Free TempList1[].
    if(TempList1!=NULL)
    {
        free(TempList1);
        TempList1 = NULL;
    }

    //Free TempList2[].
    if(TempList2!=NULL)
    {
        free(TempList2);
        TempList2 = NULL;
    }

    //Free FInfo[].
    if(FInfo!=NULL)
    {
        free(FInfo);
        FInfo = NULL;
    }

    //Free pFProd.
    if(pFProd!=NULL)
    {
        free(pFProd);
        pFProd = NULL;
    }

    //Free Soln[].
    if(Soln!=NULL)
    {
        free(Soln);
        Soln = NULL;
    }

    return;
}

//Initialize Factorial[], Iso[][], and allocate memory. Returns true if successful, false if
//it failed to allocate the memory.
bool Init()
{
    long i, j, k, p, n;

    //Deallocate any memory that is already allocated.
    Cleanup();

    //Initialize Factorial[].
    Factorial[0] = 1;
    for(i=1;i<=orderOfH;i++)
        Factorial[i] = Factorial[i-1]*i;

    //Allocate memory for Iso[][].
    Iso = (long**)malloc(sizeof(long*)*Factorial[orderOfH]);
    if(Iso==NULL)
    {
        Cleanup();
        return false;
    }

    for(i=0;i<Factorial[orderOfH];i++)
        Iso[i] = NULL;

    for(i=0;i<Factorial[orderOfH];i++)
    {
        Iso[i] = (long*)malloc(sizeof(long)*orderOfH);
        if(Iso[i]==NULL)
        {
            Cleanup();
            return false;
        }
    }

    //Initialize Iso[][].

    //Set Iso[0][] to be the identity map.
    for(n=0;n<orderOfH;n++)
        Iso[0][n] = n;

    //p = number of permutations filled in.
    p = 1; 

    //Inductively build up Iso[][]:
    //Assume {Iso[k][] : k<=i!} are all permutations which permute {0,1,...,i-1} but leave
    //{i,i+1,...,orderOfH-1}.
    //Generate new permutations by swapping i and j in Iso[k][], for all j<i and k<=i!.
    for(i=1;i<orderOfH;i++)
    for(j=0;j<i;j++)
    for(k=0;k<Factorial[i];k++)
    {
        for(n=0;n<orderOfH;n++)
        {
            if(Iso[k][n]==i)
                Iso[p][n] = j;
            else
            if(Iso[k][n]==j)
                Iso[p][n] = i;
            else
                Iso[p][n] = Iso[k][n];
        }

        p++;
    }

    //Allocate memory for Ineq[].
    Ineq = (CLargeInt*)malloc(sizeof(CLargeInt)*maxNoOfGraphs);
    if(Ineq==NULL)
    {
        Cleanup();
        return false;
    }

    //Allocate memory for H[].
    H = (CGraph*)malloc(sizeof(CGraph)*maxNoOfGraphs);
    if(H==NULL)
    {
        Cleanup();
        return false;
    }

    //Allocate memory for TempList1[].
    TempList1 = (CGraph*)malloc(sizeof(CGraph)*maxNoOfGraphs);
    if(TempList1==NULL)
    {
        Cleanup();
        return false;
    }

    //Allocate memory for TempList2[].
    TempList2 = (CGraph*)malloc(sizeof(CGraph)*maxNoOfGraphs);
    if(TempList2==NULL)
    {
        Cleanup();
        return false;
    }

    //Allocate memory for the flag information, FInfo[].
    FInfo = (CFlagInfo*)malloc(sizeof(CFlagInfo)*noOfTypes);
    if(FInfo==NULL)
    {
        Cleanup();
        return false;
    }

    //Allocate memory for the CFlagProducts object.
    pFProd = (CFlagProducts*)malloc(sizeof(CFlagProducts));
    if(pFProd==NULL)
    {
        Cleanup();
        return false;
    }

    //Allocate memory for the solution to the SDP problem.
    Soln = (CSolution*)malloc(sizeof(CSolution)*2*noOfTypes);
    if(Soln==NULL)
    {
        Cleanup();
        return false;
    }

    return true;
}

//Displays the graph g using the "<<" operator.
ostream& operator<<(ostream& out, const CGraph& g)
{
    long i, j, k;
    bool bFirst;

    //Display the order of the graph.
    out << "Order " << g.nV;
    
    out << " : ";

    //Display the edges.

    out << "Edges = {";

    //Used to help us put the commas in the right place.
    bFirst = true;

    for(i=0;i<g.nV;i++)
    for(j=0;j<i;j++)
    for(k=0;k<j;k++)
        if(g.M[k][j][i]==true)
        {
            //Display the commas that separate the edges.
            if(bFirst==false)
                out << ", ";
            else
                bFirst = false;
                
            //Display the edge ijk.
            out << k+1 << j+1 << i+1;
        }

    out << "}";

    return out;
}

//Returns the number of edges in the graph g.
long NumberOfEdges(CGraph& g)
{
    long i, j, k;
    long edges;

    edges = 0;

    for(i=0;i<g.nV;i++)
    for(j=0;j<i;j++)
    for(k=0;k<j;k++)
        if(g.M[i][j][k]==true)
            edges++;

    return edges;
}

//Returns true if g1 and g2 are of the same order and have the same edge set.
bool IsCopy(CGraph& g1, CGraph& g2)
{
    long i, j, k;

    if(g1.nV!=g2.nV)
        return false;

    for(i=0;i<g1.nV;i++)
    for(j=0;j<i;j++)
    for(k=0;k<j;k++)
        if(g1.M[i][j][k]!=g2.M[i][j][k])
            return false;

    return true;
}

//If bInduced = false, IsSubgraph() returns true if gSub is a subgraph of gSuper.
//If bInduced = true,  IsSubgraph() returns true if gSub is an induced subgraph of gSuper.
bool IsSubgraph(CGraph& gSub, CGraph& gSuper, bool bInduced)
{
    long i, j, k, p;
    bool bIsSub;

    if(gSuper.nV<gSub.nV)
        return false;

    //Permute the vertices of gSuper, to create all its isomorphic graphs, and check its edge
    //set with that of gSub. 
    for(p=0;p<Factorial[gSuper.nV];p++)
    {
        //Due to the way we constructed Iso[][],
        //{Iso[p][0], Iso[p][1], ..., Iso[p][gSuper.nV-1]} is a permutation of 
        //{0, 1, ..., gSuper.nV-1}.

        //Assume gSub is a (induced) subgraph. If at some point we've shown it isn't, then
        //try a different permutation.
        bIsSub = true;

        //Check the triple ijk (where i>j>k).
        for(i=0;i<gSub.nV;i++)
        {
            for(j=0;j<i;j++)
            {
                for(k=0;k<j;k++)
                {
                    if(bInduced==true)
                    {
                        //Check if the edge sets are identical.
                        if(gSub.M[i][j][k]!=gSuper.M[Iso[p][i]][Iso[p][j]][Iso[p][k]])
                        {
                            //Try a different permutation p.
                            bIsSub = false;
                            break;
                        }
                    }
                    else
                    {
                        //Check if the edge set of gSub is a subset.
                        if(gSub.M[i][j][k]==true
                        && gSuper.M[Iso[p][i]][Iso[p][j]][Iso[p][k]]==false)
                        {
                            //Try a different permutation p.
                            bIsSub = false;
                            break;
                        }
                    }
                }

                if(bIsSub==false)
                    break;
            }

            if(bIsSub==false)
                break;
        }

        if(bIsSub==true)
            return true;
    }

    //Tested all permutations with no success.
    return false;
}

//Each edge set of a graph can be assigned a unique integer. FindMinIso() creates gOut a
//graph with the property that (gOut, orderOfType) is isomorphic to the flag 
//(gIn, orderOfType) and has the smallest integer associated with its edge set out of all
//such graphs. We use this function to efficiently avoid storing isomorphic copies of flags. 
//This function is used in GenerateH() and GenerateFlags() via the function Generate() to
//inductively build up flags and H[]. In order for these methods to function correctly the
//way the function constructs the integer from the edge set is important.
void FindMinIso(CGraph gIn, CGraph& gOut, long orderOfType)
{
    long i, j, k, i2, j2, k2, p;
    bool bFoundSmaller, bBreak;

    //gOut holds the graph with the smallest integer associated with its edge set so far.
    gOut = gIn;

    //We wish to test all isomorphic flags of (gIn, orderOfType). We can construct them by
    //permutating vertices {orderOfType, orderOfType+1, ..., gIn.nV-1} of gIn, whilst keeping
    //vertices {0,1,...,orderOfType-1} fixed. Due to the way we constructed Iso[][], 
    //{gIn.nV-1-Iso[p][gIn.nV-1], gIn.nV-1-Iso[p][gIn.nV-2], ..., gIn.nV-1-Iso[p][0]}
    //describes just such a permutation (for 0 <= p < Factorial[gIn.nV-orderOfType]).
    for(p=0;p<Factorial[gIn.nV-orderOfType];p++)
    {
        bFoundSmaller = false;
        bBreak        = false;

        //We construct the unique integer associated with the edge set of a graph g by
        //considering an ordering of all triples ijk (with g.nV > i > j > k >=0).
        //Specifically ijk > rst if
        //   i>r,
        //or i==r and j>s,
        //or i==r and j==s and k>t.
        //Our integer is a "gIn.nV choose 3" bit number where each bit is associated with a
        //triple in our ordering. The largest triple is associated with least significant
        //bit, and the second largest triple with the second least significant bit, etc.
        //The value of the bits in the integer are given by the associated triples, with a
        //value of 1 if ijk is an edge in g and 0 if it isn't.

        //We will not explicitly construct such integers, but by examining the edge set of
        //gOut and gIn, under the permutation p, we will make a comparison of which
        //associated integer would be smaller.

        //Cycle through the triples from the smallest to the largest. Note that we can take
        //i >= orderOfType because of the way we are permuting gIn.
        for(i=orderOfType;i<gIn.nV;i++)
        {
            i2 = gIn.nV-1-Iso[p][gIn.nV-1-i];

            for(j=0;j<i;j++)
            {
                j2 = gIn.nV-1-Iso[p][gIn.nV-1-j];
                
                for(k=0;k<j;k++)
                {
                    k2 = gIn.nV-1-Iso[p][gIn.nV-1-k];

                    //Compare bits

                    if(gIn.M[i2][j2][k2]==false && gOut.M[i][j][k]==true)
                    {
                        //gIn under permutation p is smaller than gOut.
                        //Store gIn under permutation p in gOut. 
                        bFoundSmaller = true;
                        bBreak        = true;
                        break;
                    }

                    if(gIn.M[i2][j2][k2]==true && gOut.M[i][j][k]==false)
                    {
                        //gIn under permutation p is larger than gOut.
                        //Try a different value of p.
                        bFoundSmaller = false;
                        bBreak        = true;
                        break;
                    }

                    //The integers are equal so far. Try the next triple ijk.
                }

                if(bBreak==true)
                    break;
            }

            if(bBreak==true)
                break;
        }

        if(bFoundSmaller==true)
        {
            //Copy the transformed gIn into gOut
            for(i=0;i<gIn.nV;i++)
            {
                i2 = gIn.nV-1-Iso[p][gIn.nV-1-i];

                for(j=0;j<gIn.nV;j++)
                {
                    j2 = gIn.nV-1-Iso[p][gIn.nV-1-j];
                    
                    for(k=0;k<gIn.nV;k++)
                    {
                        k2 = gIn.nV-1-Iso[p][gIn.nV-1-k];

                        gOut.M[i][j][k] = gIn.M[i2][j2][k2];
                    }
                }
            }
        }

        //Check if the next permutation produces a smaller integer.
    }

    return;
}

//Used in GenerateH() and GenerateFlags() to create a list of admissible graphs/flags of a
//given order not containing isomorphic copies. Takes as input the flag
//(inputFlag, orderOfType), which it uses to create admissible flags of order inputFlag.nV+1
//that have inputFlag as an induced subgraph. It stores the new admissible graphs it finds in
//List[Found], List[Found+1], ..., etc, and updates the value of Found accordingly. Returns
//false if it reaches the end of the List[] array which it assumes is of size maxNoOfGraphs.
bool Generate(CGraph& inputFlag, long orderOfType, CGraph* List, long& Found)
{
    long   i, j, k;
    long   b, n;
    long   noOfEdgesToSet;
    bool   e;
    CGraph g, gMin;

    //Create a new graph which is inputFlag with an extra vertex v.
    g = inputFlag;
    g.nV++;

    //We need to determine which edges containing vertex v our new graph will contain.
    noOfEdgesToSet = (g.nV-1)*(g.nV-2)/2;

    //There are 2^noOfEdgesToSet possible choices for the set of edges containing v. We let b
    //encode which of the edges we include in g via the value of its bits.
    //Bits    0,   1,   2,   3,   4,   5,   6,   7,   8, ... of b tells us whether edges 
    //edges 01v, 02v, 12v, 03v, 13v, 23v, 04v, 14v, 24v, ... respectively are in graph g.
    
    //Test all sets of edges containing v.
    for(b=0;b<1<<noOfEdgesToSet;b++)
    {
        n = 0;      //n = position of bit in b. 
        i = g.nV-1; //i = vertex v.
        for(j=0;j<i;j++)
        for(k=0;k<j;k++)
        {
            //Check if ikj is in our graph, by inspecting bit n of b.
            if((b>>n)%2==1)
                e = true;
            else
                e = false;

            //Update the adjacency matrix.
            g.M[i][j][k] = e;
            g.M[i][k][j] = e;
            g.M[j][i][k] = e;
            g.M[j][k][i] = e;
            g.M[k][i][j] = e;
            g.M[k][j][i] = e;
            
            n++;
        }

        //Check if the graph is admissible.
        for(i=0;i<noOfForbiddenGraphs;i++)
            if(IsSubgraph(ForbiddenGraph[i],g,bForbidOnlyInducedSubgraphs)==true)
                break;

        if(i!=noOfForbiddenGraphs)
            continue; //A forbidden graph was found.

        //Due to the way FindMinIso() chooses a minimal representive of the isomorphic copies
        //of a flag and the way we are inductively building up our list of flags from smaller
        //order flags which are minimally represented, we are guaranteed to generate all
        //flags which are minimally represented. Hence if the flag is not already the
        //minimal representation of all its isomorphic copies we can safely ignore it.
        FindMinIso(g,gMin,orderOfType);

        if(IsCopy(g,gMin)==false)
            continue;

        //Add to List[].
        if(Found==maxNoOfGraphs)
            return false; //Ran out of space in List[].

        List[Found] = g;
        Found++;
    }

    return true;
}

//Generates the admissible graphs H[]. Returns false if the array H[], which it assumes is of
//size maxNoOfGraphs, is too small. 
bool GenerateH()
{
    long    i, j, k;
    long    FoundOld, FoundNew;
    CGraph* pOld;
    CGraph* pNew;
    CGraph* pTemp;

    //We take a list of graphs pOld[] of order n and add a vertex to each graph in the list,
    //to generate a list of graphs pNew[] of order n+1. We then make pNew[] the new pOld[]
    //list and repeat the procedure till we get the desired order of graphs.

    //pOld, pNew are pointers, allowing us to efficiently make pNew[] the new pOld[] list.
    //They point to TempList1[], TempList2[] which should have sufficient memory to hold the
    //list of graphs. 
    pOld = TempList1;
    pNew = TempList2;
    
    if(orderOfH==0)
        pNew = H; //Store the pNew[] list in H[].

    //Fill in the admissible graph of order 0.
    pNew[0].nV = 0;

    for(i=0;i<orderOfH;i++)
    for(j=0;j<orderOfH;j++)
    for(k=0;k<orderOfH;k++)
        pNew[0].M[i][j][k] = false;

    FoundNew = 1;

    for(i=1;i<=orderOfH;i++)
    {
        //Generate pNew[], a list of graphs of order i, from pOld[] a list of graphs of order
        //i-1.

        //Swap the old and new lists.
        pTemp = pOld;
        pOld  = pNew;
        pNew  = pTemp;
        
        FoundOld = FoundNew;
        FoundNew = 0; //Clear the pNew[] list.

        if(orderOfH==i)
            pNew = H; //Store pNew[] in H[].

        //For each graph in pOld[] generate graphs for pNew[] by adding an extra vertex. Note
        //that an unlabelled graph g is equivalent to the flag (g,0), allowing us to use the
        //function Generate(). 
        for(j=0;j<FoundOld;j++)
            if(Generate(pOld[j],0,pNew,FoundNew)==false)
                return false; //Ran out of space in pNew[].
    }

    noOfH = FoundNew;

    return true;
}

//Generates admissible Flags of order FlagOrder. The type is given by FI.Type. Returns false
//if the array FI.FlagList[], which it assumes is of size maxNoOfFlags, is too small.
bool GenerateFlags(CFlagInfo& FI, long FlagOrder)
{
    long    i, j;
    long    FoundOld, FoundNew;
    CGraph* pOld;
    CGraph* pNew;
    CGraph* pTemp;

    //Deal with the nonsensical case of creating a flag with FlagOrder < FI.Type.nV.
    if(FlagOrder<FI.Type.nV)
    {
        FI.noOfFlags = 0;
        return true;
    }

    //We take a list of flags pOld[] of order n and add a vertex to each flag in the list, to
    //generate a list of flags pNew[] of order n+1. We then make pNew[] the new pOld[] list
    //and repeat the procedure till we get the desired order of flags.

    //pOld, pNew are pointers, allowing us to efficiently make pNew[] the new pOld[] list.
    //They point to TempList1[], TempList2[] which should have sufficient memory to hold the
    //list of flags. 
    pOld = TempList1;
    pNew = TempList2;

    //Initialize pNew[] with the only flag of order FI.Type.nV.
    pNew[0]  = FI.Type;
    FoundNew = 1;

    for(i=FI.Type.nV+1;i<=FlagOrder;i++)
    {
        //Swap the lists.
        pTemp = pOld;
        pOld  = pNew;
        pNew  = pTemp;

        FoundOld = FoundNew;
        FoundNew = 0; //Clear the pNew[] list.

        //For each flag in pOld[] generate flags for pNew[] by adding an extra vertex. 
        for(j=0;j<FoundOld;j++)
            if(Generate(pOld[j],FI.Type.nV,pNew,FoundNew)==false)
                return false; //Ran out of space in pNew[].
    }

    //Copy pNew[] into FI.FlagList[].
    if(FoundNew>maxNoOfFlags)
        return false; //Not enough space in FI.FlagList[].

    for(i=0;i<FoundNew;i++)
        FI.FlagList[i] = pNew[i];

    FI.noOfFlags = FoundNew;

    return true;
}

//Uses FI.FlagList[] and H[] to fill in pFProd->Data[][][].
void CalculateFlagProducts(CFlagInfo& FI)
{
    long   i, j, k, n, p, r;
    long   orderOfFlag;
    long   orderOfType;
    long   mapping[2][orderOfH];
    long   index[2];
    CGraph f[2];

    if(FI.noOfFlags<=0)
        return;

    orderOfType = FI.Type.nV;
    orderOfFlag = FI.FlagList[0].nV;

    //Clear pFProd->Data[][][].
    for(i=0;i<FI.noOfFlags;i++)
    for(j=0;j<FI.noOfFlags;j++)
    for(k=0;k<noOfH;k++)
        pFProd->Data[i][j][k] = 0;

    //Use Iso[p][] to choose vertices which will be part of our flags.
    for(p=0;p<Factorial[orderOfH];p++)
    {
        //Create a mapping from Iso[p][]. Vertex v in the flag f[r] is the vertex
        //"mapping[r][v]" in H[].

        //The induced type of the flag are given by vertices 0, 1, ..., orderOfType-1, which
        //will be vertices Iso[p][0], Iso[p][1], ..., Iso[p][orderOfType-1], in H[].
        for(i=0;i<orderOfType;i++)
        {
            mapping[0][i] = Iso[p][i];
            mapping[1][i] = Iso[p][i];
        }

        //Map the remaining vertices of the flags.
        for(i=orderOfType;i<orderOfFlag;i++)
        {
            //The vertices orderOfType, orderOfType+1, ..., orderOfFlag-1 of f[0] are the
            //vertices Iso[p][orderOfType], Iso[p][orderOfType+1], ..., Iso[p][orderOfFlag-1]
            //in H[].
            mapping[0][i] = Iso[p][i];

            //The vertices orderOfType, orderOfType+1, ..., orderOfFlag-1 of f[1] are the
            //vertices Iso[p][orderOfFlag], Iso[p][orderOfFlag+1], ...,
            //Iso[p][2*orderOfFlag-orderOfType-1] in H[].
            mapping[1][i] = Iso[p][i+orderOfFlag-orderOfType];
        }

        //Apply mapping[][] to each H[n] to get two flags f[0], f[1].
        for(n=0;n<noOfH;n++)
        {
            //Create the two flags.
            for(r=0;r<2;r++)
            {
                f[r].nV = orderOfFlag;

                for(i=0;i<orderOfFlag;i++)
                for(j=0;j<orderOfFlag;j++)
                for(k=0;k<orderOfFlag;k++)
                    f[r].M[i][j][k] = H[n].M[mapping[r][i]][mapping[r][j]][mapping[r][k]];
            }

            //Check types match, by temporarily reducing the vertex set of f[0].
            f[0].nV = orderOfType;
            if(IsCopy(f[0],FI.Type)==false)
                continue;

            f[0].nV = orderOfFlag;

            //Find the isomorphic copy of each flag in FI.FlagList[], and store the position
            //at which it appears in index[].
            for(r=0;r<2;r++)
            {
                FindMinIso(f[r],f[r],orderOfType);

                for(index[r]=0;index[r]<FI.noOfFlags;index[r]++)
                    if(IsCopy(f[r],FI.FlagList[index[r]])==true)
                        break; //Found.
            }

            //Update pFProd->Data[][][].
            pFProd->Data[index[0]][index[1]][n]++;
        }
    }

    return;
}

//Generates a new basis of flags from the information in FI. It stores the necessary data in
//FI.BasisSize[], FI.BasisElementSize[][], and FI.BasisElementFlags[][][].
void GenerateBasis(CFlagInfo& FI)
{
    long   i, j, k, m, n, p;
    long   minFlag[maxNoOfFlags];
    CGraph g;

    //Generate the orbit of each flag under the automorphism group of FI.Type, and store the
    //flag with the smallest index in minFlag[].

    //To start with we assume the unmodified flag has the smallest index out of the flags
    //that form its orbit.
    for(i=0;i<FI.noOfFlags;i++)
        minFlag[i] = i;

    //Find an element of the automorphism group, i.e. a permutation p that does not change
    //FI.Type.
    for(p=0;p<Factorial[FI.Type.nV];p++)
    {
        //Check p leaves FI.Type unchanged.
        g.nV = FI.Type.nV;
        
        for(i=0;i<FI.Type.nV;i++)
        for(j=0;j<FI.Type.nV;j++)
        for(k=0;k<FI.Type.nV;k++)
            g.M[i][j][k] = FI.Type.M[Iso[p][i]][Iso[p][j]][Iso[p][k]];

        if(IsCopy(g,FI.Type)==false)
            continue; //FI.Type under p is not FI.Type.

        //Apply p to all flags to create a member of their orbit.
        for(n=0;n<FI.noOfFlags;n++)
        {
            //Apply p to the flag.
            g.nV = FI.FlagList[n].nV;
                    
            for(i=0;i<g.nV;i++)
            for(j=0;j<g.nV;j++)
            for(k=0;k<g.nV;k++)
                g.M[i][j][k] = FI.FlagList[n].M[Iso[p][i]][Iso[p][j]][Iso[p][k]];

            //Find g in FI.FlagList[].
            FindMinIso(g,g,FI.Type.nV);

            for(m=0;m<FI.noOfFlags;m++)
                if(IsCopy(g,FI.FlagList[m])==true)
                    break;

            //Check if this new flag has a smaller index.
            if(m<minFlag[n])
                minFlag[n] = m;
        }
    }

    //Fill in the basis data.
    FI.BasisSize[0] = 0;
    FI.BasisSize[1] = 0;

    for(i=0;i<FI.noOfFlags;i++)
    {
        //Find the orbit in which the flag with the smallest index is i (if such an orbit
        //exists).
        
        if(minFlag[i]!=i)
            continue; //No such orbit.

        //Found a new invariant basis element i.e. the sum of the flags with minFlag[] = i.
        //The first flag in this element is FI.FlagList[i].
        FI.BasisElementSize[0][FI.BasisSize[0]]     = 1;
        FI.BasisElementFlags[0][FI.BasisSize[0]][0] = i;
        FI.BasisSize[0]++;

        //Find the other flags in the orbit.
        for(j=i+1;j<FI.noOfFlags;j++)
        {
            if(minFlag[j]!=i)
                continue; //Not part of the orbit.

            //Add the flag FI.FlagList[j] to the invariant element.
            m = FI.BasisElementSize[0][FI.BasisSize[0]-1];
            FI.BasisElementSize[0][FI.BasisSize[0]-1]++;
            FI.BasisElementFlags[0][FI.BasisSize[0]-1][m] = j;

            //Add a new element which is not invariant.
            FI.BasisElementSize[1][FI.BasisSize[1]] = 2;
            FI.BasisElementFlags[1][FI.BasisSize[1]][0] = j;
            FI.BasisElementFlags[1][FI.BasisSize[1]][1] = i;
            FI.BasisSize[1]++;
        }
    }

    return;
}

//Used to compare two large integers.
//Returns -1 if inputA < inputB,
//         0 if inputA = inputB,
//         1 if inputA > inputB.
long LI_Compare(CLargeInt& inputA, CLargeInt& inputB)
{
    long i;

    if(inputA.bPositive!=inputB.bPositive)
    {
        //Check if inputA = inputB = 0.
        for(i=0;i<LI_MAX;i++)
            if(inputA.Data[i]!=0 || inputB.Data[i]!=0)
                break;
        
        if(i==LI_MAX)
            return 0; //inputA = inputB = 0.

        if(inputA.bPositive==true)
            return 1;  //inputA > inputB.
        else
            return -1; //inputA < inputB.
    }

    //inputA and inputB have the same sign.

    //Check the magnitude of the bytes, beginning with the most significant byte. 
    for(i=LI_MAX-1;i>=0;i--)
    {
        if(inputA.Data[i]>inputB.Data[i])
        {
            if(inputA.bPositive==true)
                return 1;  //inputA > inputB.
            else
                return -1; //inputA < inputB.
        }

        if(inputA.Data[i]<inputB.Data[i])
        {
            if(inputA.bPositive==true)
                return -1; //inputA < inputB.
            else
                return 1;  //inputA > inputB.
        }

        //inputA.Data[i] = inputB.Data[i]. Check the next most significant byte.
    }

    return 0; //inputA = inputB.
}

//Assigns "input" to the large integer "output". Returns false if overflow occurs, which
//should never happen if LI_MAX >= 4.
bool LI_Assign(long input, CLargeInt& output)
{
    long          i;
    unsigned char c;

    //Calculate the sign of input.
    if(input<0)
    {
        //Make the input positive.
        input = -input;
        output.bPositive = false;
    }
    else
        output.bPositive = true;

    //Clear Data[].
    for(i=0;i<LI_MAX;i++)
        output.Data[i] = 0;

    //Initialize Data[] so it equals input.
    for(i=0;i<sizeof(long);i++)
    {
        //Remove the least significant byte of input and store it in c.
        c     = input%256;
        input = input/256;

        //Store c in Data[].

        if(c==0)
            continue;

        if(i>=LI_MAX)
            return false; //Ran out of space in Data[]. Need to increase LI_MAX.

        output.Data[i] = c;
    }

    return true;
}

//Assigns inputA+inputB to output. Returns false if overflow occurs.
bool LI_Add(CLargeInt inputA, CLargeInt inputB, CLargeInt& output)
{
    long i;
    long sum, carryOver;

    if(inputA.bPositive==inputB.bPositive)
    {
        //Add the magnitudes of inputA and inputB.

        output.bPositive = inputA.bPositive;

        //Do the addition byte by byte.
        carryOver = 0;
        for(i=0;i<LI_MAX;i++)
        {
            sum            = inputA.Data[i]+inputB.Data[i]+carryOver;
            output.Data[i] = sum%256;
            carryOver      = sum/256;
        }

        if(carryOver!=0)
            return false; //Overflow occured. Need to increase LI_MAX.

        return true;
    }
    else
    {
        //Find the difference between the magnitudes of inputA and inputB.

        //Assume the magnitude of inputA is greater than or equal to that of inputB.
        output.bPositive = inputA.bPositive;

        //Do the subtraction of inputB from inputA, byte by byte.
        carryOver = 0;
        for(i=0;i<LI_MAX;i++)
        {
            sum = inputA.Data[i]-inputB.Data[i]+carryOver;
            
            if(sum>=0)
            {
                output.Data[i] = sum;
                carryOver      = 0;
            }
            else
            {
                output.Data[i] = 256+sum;
                carryOver      = -1;
            }
        }

        if(carryOver==0)
            return true;

        //The carryOver is non-zero hence the magnitude of inputB must be bigger than
        //inputA.
        output.bPositive = inputB.bPositive;

        //Do the subtraction of inputA from inputB, byte by byte.
        carryOver = 0;
        for(i=0;i<LI_MAX;i++)
        {
            sum = inputB.Data[i]-inputA.Data[i]+carryOver;
            
            if(sum>=0)
            {
                output.Data[i] = sum;
                carryOver      = 0;
            }
            else
            {
                output.Data[i] = 256+sum;
                carryOver      = -1;
            }
        }

        return true;
    }

    //Should never reach this point, but just in case.
    return false;
}

//Assigns inputA-inputB to output. Returns false if overflow occurs.
bool LI_Subtract(CLargeInt inputA, CLargeInt inputB, CLargeInt& output)
{
    inputB.bPositive = !inputB.bPositive;
    return LI_Add(inputA,inputB,output);
}

//Assigns inputA*inputB to output. Returns false if overflow occurs.
bool LI_Multiply(CLargeInt inputA, CLargeInt inputB, CLargeInt& output)
{
    long      i, j;
    long      product;
    CLargeInt Temp;

    //We will first find the product of the magnitudes of inputA and inputB and store it in
    //output. Temp is used to temporarily store results of intermediate calculations.
    
    //Initialize output and Temp to 0.
    output.bPositive = true;
    Temp.bPositive   = true;

    for(i=0;i<LI_MAX;i++)
    {
        output.Data[i] = 0;
        Temp.Data[i]   = 0;
    }

    //Multiply byte by byte.
    for(i=0;i<LI_MAX;i++)
    for(j=0;j<LI_MAX;j++)
    {
        //Set Temp = inputA.Data[i]*(256^i)*inputB.Data[j]*(256^j), and add it to output.

        product = inputA.Data[i]*inputB.Data[j];
        
        if(product==0)
            continue;

        //Temp = product*(256^(i+j)).

        //Check whether product fits into one byte or two.
        if(product/256==0)
        {
            //product fits into one byte.

            //Check if Temp would be too large to fit in CLargeInt.
            if(i+j>=LI_MAX)
                return false; //Overflow occured. Need to increase LI_MAX.

            //Set Temp.
            Temp.Data[i+j] = product;

            //output += Temp.
            if(LI_Add(output,Temp,output)==false)
                return false; //Overflow occured. Need to increase LI_MAX.

            //Set Temp back to 0.
            Temp.Data[i+j] = 0;
        }
        else
        {
            //product requires two bytes.

            //Check if Temp would be too large to fit in CLargeInt. 
            if(i+j+1>=LI_MAX)
                return false; //Overflow occured. Need to increase LI_MAX.

            //Set Temp.
            Temp.Data[i+j]   = product%256;
            Temp.Data[i+j+1] = product/256;

            //output += Temp.
            if(LI_Add(output,Temp,output)==false)
                return false; //Overflow occured. Need to increase LI_MAX.

            //Set Temp back to 0.
            Temp.Data[i+j]   = 0;
            Temp.Data[i+j+1] = 0;
        }
    }

    //Calculate the sign of inputA*inputB.
    if(inputA.bPositive==inputB.bPositive)
        output.bPositive = true;
    else
        output.bPositive = false;

    return true;
}

//We define |input| to be the magnitude of input, and sign(input) to be 1 if input >= 0 and
//-1 if input < 0. The function calculates outputQ, and outputR as follows:
//outputQ = sign(inputN)*sign(inputD)*(|inputN|/|inputD|), where '/' is integer division.
//outputR = sign(inputN)*(|inputN|%|inputD|).
//The choice of signs for outputQ and outputR were chosen so that
//inputN = inputD*outputQ + outputR holds. Returns false if inputD = 0 or if outputQ resides
//in the same address as outputR.
bool LI_Divide(CLargeInt inputN, CLargeInt inputD, CLargeInt& outputQ, CLargeInt& outputR)
{
    long      i, j, k;
    long      firstN, firstD; //The positions of the most significant bits.
    long      lastTemp;
    long      carryOver;
    bool      bSubtractD;
    CLargeInt temp;

    //Check if outputQ and outputR reside in the same memory address.
    if(&outputQ==&outputR)
        return false;

    //Calculate the signs of outputQ and outputR.
    if(inputN.bPositive==inputD.bPositive)
        outputQ.bPositive = true;
    else
        outputQ.bPositive = false;

    outputR.bPositive = inputN.bPositive;

    //We will do the division via the long division method in binary.

    //Find the leading bit of inputD.

    //Find the most significant non-zero byte.
    for(i=LI_MAX-1;i>=0;i--)
        if(inputD.Data[i]!=0)
            break;

    if(i==-1)
        return false; //All entries were zero, hence inputD = 0.

    //Find the most significant non-zero bit.
    for(j=7;j>=0;j--)
        if((inputD.Data[i]>>j)%2==1)
            break;

    //Store the leading bit.
    firstD = 8*i+j;

    //Find the leading bit of inputN.

    //Find the most significant non-zero byte.
    for(i=LI_MAX-1;i>=0;i--)
        if(inputN.Data[i]!=0)
            break;

    if(i==-1)
    {
        //All entries were zero, hence inputN = 0. Set outputQ = outputR = 0.
        for(i=0;i<LI_MAX;i++)
        {
            outputQ.Data[i] = 0;
            outputR.Data[i] = 0;
        }

        return true;
    }

    //Find the most significant non-zero bit.
    for(j=7;j>=0;j--)
        if((inputN.Data[i]>>j)%2==1)
            break;

    //Store the leading bit.
    firstN = 8*i+j;

    if(firstD>firstN)
    {
        //inputD is of greater magnitude than inputN.
        //Set outputQ = 0.
        //Set outputR = inputN.
        for(i=0;i<LI_MAX;i++)
        {
            outputQ.Data[i] = 0;
            outputR.Data[i] = inputN.Data[i];
        }

        return true;
    }
    
    //Create an integer "temp" comprised of the leading firstD+1 bits of inputN.
    temp.bPositive = true;
    for(i=0;i<LI_MAX;i++)
        temp.Data[i] = 0;

    for(i=0;i<firstD+1;i++)
    {
        //Position of bit in inputN.
        j = firstN-firstD+i;
        
        //The value of the bit.
        k = (inputN.Data[j/8]>>j%8)%2;
        
        //Store the bit in temp.
        temp.Data[i/8] |= k<<i%8;
    }

    //Store the position in inputN of the least significant bit that makes up temp.
    lastTemp = firstN-firstD;

    //Initialize outputQ to 0.
    for(i=0;i<LI_MAX;i++)
        outputQ.Data[i] = 0;

    inputD.bPositive = true;

    //Calculate the value of the bits of outputQ.
    while(true)
    {
        //Calculate the value of the lastTemp bit in outputQ.

        //Check if temp >= inputD, if so we subtract inputD from temp.
        for(i=LI_MAX-1;i>=0;i--)
        {
            if(temp.Data[i]>inputD.Data[i])
            {
                //temp > inputD.
                bSubtractD = true;
                break;
            }

            if(temp.Data[i]<inputD.Data[i])
            {
                //temp < inputD.
                bSubtractD = false;
                break;
            }
        }

        if(i==-1)
            bSubtractD = true; //temp = inputD.

        if(bSubtractD==true)
        {
            //Subtract inputD from temp.
            carryOver = 0;
            for(i=0;i<LI_MAX;i++)
            {
                j = temp.Data[i]-inputD.Data[i]+carryOver;
                if(j>=0)
                {
                    temp.Data[i] = j;
                    carryOver    = 0;
                }
                else
                {
                    temp.Data[i] = 256+j;
                    carryOver    = -1;
                }
            }

            //Set the lastTemp bit of outputQ to 1.
            outputQ.Data[lastTemp/8] |= 1<<lastTemp%8; 
        }

        if(lastTemp==0)
            break; //Found all the bits of outputQ.

        //Multiply temp by 2 and append the "lastTemp-1" bit from inputN to temp.
        lastTemp--;
        carryOver = (inputN.Data[lastTemp/8]>>lastTemp%8)%2;
        
        for(i=0;i<LI_MAX;i++)
        {
            j            = 2*temp.Data[i]+carryOver;
            carryOver    = j/256;
            temp.Data[i] = j%256;
        }

        //Calculate the next bit of outputQ.
    }

    //temp holds the remainder. Set outputR = temp.
    for(i=0;i<LI_MAX;i++)
        outputR.Data[i] = temp.Data[i];

    return true;
}

//Displays the large integer L using the "<<" operator.
ostream& operator<<(ostream& out, const CLargeInt& L)
{
    long      i, n;
    char      digit[3*LI_MAX];
    CLargeInt Q, R, D, Z;

    //We represent 0 and 10 as CLargeInt objects Z and D respectively.
    LI_Assign(0,Z);
    LI_Assign(10,D);

    //Set Q to be initially the magnitude of L.
    Q = L;
    Q.bPositive = true;

    //Remove the least significant decimal digit of Q and store it in digit[].
    n = 0;
    while(LI_Compare(Q,Z)!=0)
    {
        //R = Q%10.
        //Q = Q/10.
        LI_Divide(Q,D,Q,R);
        digit[n] = R.Data[0];
        n++;
    }

    //Check if L = 0.
    if(n==0)
    {
        out << "0";
        return out;
    }

    //Display the sign of the integer.
    if(L.bPositive==false)
        out << "-";

    //Display the digits of the integer.
    for(i=n-1;i>=0;i--)
        out << (char)('0'+digit[i]);

    return out;
}

//Reads a large integer L from the input stream "in". It expects the integer to be terminated
//by a white space character (' ', '\n', '\r', '\t', or the end of file marker).
//Returns  1 on success,
//        -1 if the input is formatted incorrectly,
//        -2 if overflow occured.
long ReadInLargeInt(istream& in, CLargeInt& L)
{
    char      c;
    CLargeInt A;
    bool      bIsPositive;
    bool      bReadDigit;

    LI_Assign(0,L);

    if(in.eof()==true)
        return -1; //The input is empty.

    //Ignore white space characters at the start of the integer.
    while(true)
    {
        in.get(c);

        if(in.eof()==true)
            return -1; //The input ends before a digit could be read.

        //Check if the character is not a white space.
        if(c!=' ' && c!='\t' && c!='\n' && c!='\r')
            break;
    }

    //bReadDigit keeps track of whether we have read at least one digit.
    bReadDigit = false;

    if(c=='+')
        bIsPositive = true;
    else
    if(c=='-')
        bIsPositive = false;
    else
    if('0'<=c && c<='9')
    {
        bIsPositive = true;

        //c-'0' is the numerical digit which c represents.
        LI_Assign(c-'0',L);
        bReadDigit = true;
    }
    else
        return -1; //An unexpected character was encountered.

    //Append digits to L until we reach a white space.
    while(true)
    {
        in.get(c);

        if(bReadDigit==true)
        {
            //Check if we have reached the end of the integer. If so set its sign and return
            //1 to indicate success.

            if(in.eof()==true)
            {
                L.bPositive = bIsPositive;
                return 1;
            }

            if(c==' ' || c=='\t' || c=='\n' || c=='\r')
            {
                L.bPositive = bIsPositive;
                return 1;
            }
        }
        else
        {
            if(in.eof()==true)
                return -1; //The input ends before a digit could be read.
        }

        if('0'<=c && c<='9')
        {
            bReadDigit = true;

            //Append the digit to L, by setting L = 10*L + c-'0'.
            LI_Assign(10,A);

            if(LI_Multiply(L,A,L)==false)
                return -2; //Overflow occured.

            LI_Assign(c-'0',A);

            if(LI_Add(L,A,L)==false)
                return -2; //Overflow occured.
        }
        else
            return -1; //An unexpected character was encountered.
    }

    //We should never reach this point, but just in case return -1.
    return -1;
}

//This function uses the floating point entries of the symmetric matrix given by
//soln.Matrix[][] to initialize soln.LMatrix[]. We use soln.LMatrix[][] to construct
//(Lmatrix^T LMatrix)/4^accuracy which is a positive semidefinite matrix that is a rational
//approximation of soln.Matrix[][]. The function modifies the entries of soln.Matrix[][],
//and so the data it originally held will be lost. 
bool RationalizeSolution(CSolution& soln)
{
    long           i, j, k, d;
    long           dim;
    double         multiplier;
    double         Track[maxNoOfFlags][maxNoOfFlags];
    unsigned char* pByte;
    unsigned char  byte[8];
    bool           isBigEndian;
    long           signBit;
    long           exponent;
    CLargeInt      Temp;

    //Our aim is to first diagonalize the symmetric matrix soln.Matrix[][]. We will achieve
    //this by doing symmetric row and column operations which must necessarily leave the
    //matrix symmetric. We will keep track of these operations via the matrix
    //Track[][] so that the "original soln.Matrix" 
    //= Track^T * "modified soln.Matrix" * Track. Once soln.Matrix[][] is
    //diagonalised the entries of Track[][] are approximated as integers over 2^accuracy
    //which are stored in soln.LMatrix[][].

    dim = soln.dim;

    //Initialize Track[][] to be the identity matrix.
    for(i=0;i<dim;i++)
    for(j=0;j<dim;j++)
    {
        if(i==j)
            Track[i][j] = 1;
        else
            Track[i][j] = 0;
    }

    //Start diagonalizing soln.Matrix[][].
    for(d=0;d<dim;d++)
    {
        //Make all entries in row d and column d zero, save for the soln.Matrix[d][d] entry.

        //To avoid division by zero we require soln.Matrix[d][d] to be non-zero. We use
        //"tolerance" to avoid values that are close to zero.
        if(-tolerance<=soln.Matrix[d][d] && soln.Matrix[d][d]<=tolerance)
        {
            //soln.Matrix[d][d] is zero, try to do row/column operations to make it non-zero.

            //Find a non-zero element in row d.
            for(i=d+1;i<dim;i++)
                if(soln.Matrix[d][i]<-tolerance || tolerance<soln.Matrix[d][i])
                    break;

            if(i==dim)
            {
                //All entries are 0.
                
                //Force the entries to be 0 to remove rounding error.
                for(i=d;i<dim;i++)
                {
                    soln.Matrix[d][i] = 0;
                    soln.Matrix[i][d] = 0;
                }

                //Make soln.Matrix[d][d] = 1, and update Track[][] accordingly.
                soln.Matrix[d][d] = 1;

                for(j=0;j<dim;j++)
                    Track[d][j] = 0;

                continue;
            }
            
            //Repeatedly add row i to row d and column i to column d until soln.Matrix[d][d]
            //is non-zero.
            while(-tolerance<=soln.Matrix[d][d] && soln.Matrix[d][d]<=tolerance)
            {
                for(j=d;j<dim;j++)
                    soln.Matrix[d][j] += soln.Matrix[i][j]; 

                for(j=d;j<dim;j++)
                    soln.Matrix[j][d] += soln.Matrix[j][i];

                //Update Track[][] by subtracting row d from row i.
                for(j=0;j<dim;j++)
                    Track[i][j] -= Track[d][j];
            }
        }

        //soln.Matrix[d][d] is non-zero.

        //Subtract row d from all other rows so that entries in column d are zero. Subtract
        //column d from the other columns so that entries in row d are zero.

        //We update Track[][] first as it uses information from soln.Matrix[][] that
        //will be destroyed.
        for(i=d+1;i<dim;i++)
        {
            //Add (soln.Matrix[i][d]/soln.Matrix[d][d])*row i to row d.
            for(j=0;j<dim;j++)
                Track[d][j] += Track[i][j]*soln.Matrix[i][d]/soln.Matrix[d][d];
        }

        //Subtract the rows.
        for(i=d+1;i<dim;i++)
        {
            //Subtract (soln.Matrix[i][d]/soln.Matrix[d][d])*row d from row i.
            multiplier = soln.Matrix[i][d]/soln.Matrix[d][d];

            for(j=d;j<dim;j++)
                soln.Matrix[i][j] -= multiplier*soln.Matrix[d][j];
        }

        //Subtract the columns. Because column d is zero (except for soln.Matrix[d][d]) this
        //is equivalent to setting row d to zero.
        for(i=d+1;i<dim;i++)
            soln.Matrix[d][i] = 0;


        //Make soln.Matrix[d][d] = 1, and update Track[][] by multiplying row d by the
        //square root of Matrix[d][d].

        if(soln.Matrix[d][d]<=0)
        {
            //The matrix is not positive semidefinite. To remedy this we force
            //soln.Matrix[d][d] to be 0.
            soln.Matrix[d][d] = 0;
        }

        multiplier = sqrt(soln.Matrix[d][d]);

        for(j=0;j<dim;j++)
            Track[d][j] *= multiplier;

        soln.Matrix[d][d] = 1;
    }

    //Approximate the entry Track[i][j] by the rational number
    //soln.LMatrix[i][j]/(2^accuracy) for an appropriate choice of soln.LMatrix[i][j]. We
    //assume the type double is 8 bytes long and is equivalent to binary64 as defined in the
    //IEEE 754-2008 standard.

    //Check endianness (i.e. the order bytes are stored in multibyte types).
    k     = 1;
    pByte = (unsigned char*)(&k);
    
    if(pByte[0]==0)
        isBigEndian = true;
    else
        isBigEndian = false;
        

    for(i=0;i<dim;i++)
    for(j=0;j<dim;j++)
    {
        //Convert Track[i][j] into an array of 8 bytes stored in little-endian order.
        pByte = (unsigned char*)(&Track[i][j]);
        
        if(isBigEndian==true)
        {
            //Reverse the byte order.
            for(k=0;k<8;k++)
                byte[k] = pByte[7-k]; 
        }
        else
        {
            for(k=0;k<8;k++)
                byte[k] = pByte[k];
        }

        //The most significant bit stores the sign:
        //0 = positive,
        //1 = negative.
        signBit = byte[7]>>7; 

        //The next 11 most significant bits store the exponent offset by 1023.
        exponent = (byte[7]&127)<<4 | byte[6]>>4;

        if(exponent==2047)
            return false; //Track[i][j] is infinity or not a number.
        
        if(exponent==0)
        {
            //Initialize soln.LMatrix[i][j] with the highest 4 bits of the fractional part,
            //do not include the "unstored most significant bit".
            LI_Assign(byte[6]&15,soln.LMatrix[i][j]);

            //Offset the exponent. (There are 52 bits stored in the fractional part.)
            exponent = -1023+1-52;
        }
        else
        {
            //Initialize soln.LMatrix[i][j] with the highest 4 bits of the fractional part,
            //include the unstored most significant bit.
            LI_Assign(byte[6]&15 | 1<<4,soln.LMatrix[i][j]);

            //Offset the exponent. (There are 52 bits stored in the fractional part.)
            exponent = exponent-1023-52;
        }

        //Fill in the rest of the fractional part.
        for(k=5;k>=0;k--)
        {
            LI_Assign(256,Temp);
            if(LI_Multiply(Temp,soln.LMatrix[i][j],soln.LMatrix[i][j])==false)
                return false; //Overflow occured.

            LI_Assign(byte[k],Temp);
            if(LI_Add(Temp,soln.LMatrix[i][j],soln.LMatrix[i][j])==false)
                return false; //Overflow occured.
        }

        if(signBit==1)
            soln.LMatrix[i][j].bPositive = false;

        //Currently Track[i][j] =  soln.LMatrix[i][j] * 2^exponent
        //= (soln.LMatrix[i][j] * 2^(exponent+accuracy))/2^accuracy. Hence we want
        //soln.LMatrix[i][j] to be soln.LMatrix[i][j] * 2^(exponent+accuracy).
        
        exponent += accuracy;

        while(exponent<0)
        {
            LI_Assign(2,Temp);
            LI_Divide(soln.LMatrix[i][j],Temp,soln.LMatrix[i][j],Temp);
            exponent++;
        }

        while(exponent>0)
        {
            LI_Assign(2,Temp);
            if(LI_Multiply(Temp,soln.LMatrix[i][j],soln.LMatrix[i][j])==false)
                return false;
            exponent--;
        }
    }

    //Because Track[][] is (approximately) soln.LMatrix[][] / 2^accuracy, our original
    //soln.Matrix is (approximately) (soln.LMatrix^T * soln.LMatrix) / 4^accuracy.

    return true;
}

//Creates the ".soln" file from the output of the SDP solver. It is called by main() and
//expects FInfo[].BasisSize[] to be already computed. Returns false if some sort of error was
//encountered, and true if it was successful.
bool CreateSolnFile()
{
    long     i, j, k, n;
    long     nBlocks;
    double   d;
    double   maxEntries[noOfTypes];
    ifstream FileIn;
    ofstream FileOut;

    //Create the ".soln" file.
            
    //Read in the floating point solution to the SDP problem. Expects the file to contain the
    //values of the variables, in order, separated by spaces.
    cout << "Reading in the solution......" << flush;
    
    FileIn.open(filenameOutput);

    if(FileIn.good()==false)
    {
        cout << endl << endl;
        cout << "ERROR: Cannot open the file \"" << filenameOutput << "\"." << endl;
        return false;
    }

    //The first variable is the density, which we ignore.
    FileIn >> d;

    //Initialize Soln[].
    nBlocks = 0;
    for(n=0;n<noOfTypes;n++)
    {
        //Keep track of the largest magnitudes of matrix entries. Which is useful for
        //identifying non-contributing types.
        maxEntries[n] = 0;

        for(k=0;k<2;k++)
        {
            if(FInfo[n].BasisSize[k]<=0)
                continue; //Ignore blocks of dimension 0.

            Soln[nBlocks].dim = FInfo[n].BasisSize[k];

            for(i=0;i<Soln[nBlocks].dim;i++)
            for(j=i;j<Soln[nBlocks].dim;j++)
            {
                FileIn >> d;

                Soln[nBlocks].Matrix[i][j] = d;
                Soln[nBlocks].Matrix[j][i] = d;

                if(d<0)
                    d = -d;

                if(d>maxEntries[n])
                    maxEntries[n] = d;
            }

            nBlocks++;
        }
    }

    FileIn.close();

    cout << "Done." << endl;

    //Display the largest magnitude of the entries for each type.
    cout << "Largest value in matrix:" << endl;
    for(n=0;n<noOfTypes;n++)
        cout << "Type[" << n << "] : " << maxEntries[n] << endl;

    cout << endl;

    //Rationalize the solution.
    cout << "Rationalizing the solution..." << flush;

    for(n=0;n<nBlocks;n++)
        if(RationalizeSolution(Soln[n])==false)
        {
            cout << endl << endl;
            cout << "ERROR: Try changing LI_MAX, accuracy, and tolerance." << endl;
            return false;
        }

    cout << "Done." << endl;
    cout << endl;

    //Write the ".soln" file.
    cout << "Writing the \".soln\" file....." << flush;

    FileOut.open(filenameSoln,ios::trunc);
            
    if(FileOut.good()==false)
    {
        cout << endl << endl;
        cout << "ERROR: Cannot create the file \"" << filenameSoln << "\"." << endl;
        return false;
    }

    //Write out the accuracy.
    FileOut << accuracy;
    
    //Write out the matrix entries.
    for(n=0;n<nBlocks;n++)
    for(i=0;i<Soln[n].dim;i++)
    for(j=0;j<Soln[n].dim;j++)
        FileOut << " " << Soln[n].LMatrix[i][j];

    FileOut << flush;

    FileOut.close();

    cout << "Done." << endl;
    cout << endl;

    return true;
}

//Initializes Soln[].LMatrix[][] using the precomputed values stored in a ".soln" file. It 
//overwrites the default value of "accuracy". The function is called by main() and expects
//FInfo[].BasisSize[] to be already computed. Returns false if some sort of error was
//encountered, and true if it was successful.
bool ReadSolnFile()
{
    long     i, j, k, n;
    long     nBlocks;
    long     ret;
    ifstream FileIn;

    //Read the ".soln" file.

    //Read in the precomputed solution. Expects the file to contain the values of the
    //variables, separated by spaces.
    cout << "Reading the \".soln\" file....." << flush;
    
    FileIn.open(filenameSoln);

    if(FileIn.good()==false)
    {
        cout << endl << endl;
        cout << "ERROR: Cannot open the file \"" << filenameSoln << "\"." << endl;
        return false;
    }

    //The first variable is the "accuracy".
    accuracy = 0;
    FileIn >> accuracy;

    if(accuracy<0)
    {
        cout << endl << endl;
        cout << "ERROR: Unexpected data encountered." << endl;
        return false;
    }

    //Initialize Soln[].LMatrix[][].
    nBlocks = 0;
    for(n=0;n<noOfTypes;n++)
    {
        for(k=0;k<2;k++)
        {
            if(FInfo[n].BasisSize[k]<=0)
                continue; //Ignore blocks of dimension 0.

            Soln[nBlocks].dim = FInfo[n].BasisSize[k];

            for(i=0;i<Soln[nBlocks].dim;i++)
            for(j=0;j<Soln[nBlocks].dim;j++)
            {
                ret = ReadInLargeInt(FileIn,Soln[nBlocks].LMatrix[i][j]);

                if(ret==-1)
                {
                    //The data is formatted incorrectly.
                    cout << endl << endl;
                    cout << "ERROR: Unexpected data encountered." << endl;
                    return false;
                }

                if(ret==-2)
                {
                    //Overflow occured.
                    cout << endl << endl;
                    cout << "ERROR: Overflow occured. Try changing LI_MAX." << endl;
                    return false;
                }
            }

            nBlocks++;
        }
    }

    FileIn.close();

    cout << "Done." << endl;
    cout << endl;

    return true;
}

//Creates a ".dat-s" file that can be used as input to a SDP solver. The function is called
//by main() and expects FInfo[] and H[] to be already computed. Returns false if some sort of
//error was encountered, and true if it was successful.
bool CreateSDPFile()
{
    long     i, j, k, m, n;
    long     b, v;
    long     r, c;
    long     nVar, nBlocks;
    long     coeff;
    long     product[maxNoOfGraphs];
    ofstream FileOut;

    //Create the ".dat-s" file for an SDP solver.

    //Calculate the number of blocks and the number of variables we are optimizing over.
    nBlocks = 0;
    nVar    = 0;

    for(i=0;i<noOfTypes;i++)
    for(j=0;j<2;j++)
        if(FInfo[i].BasisSize[j]>0)
        {
            nBlocks++;

            //We only need to determine the upper right half of the positive semidefinite
            //matrices as they are symmetric.
            nVar += FInfo[i].BasisSize[j]*(FInfo[i].BasisSize[j]+1)/2;
        }

    //We have an extra variable $\rho$, that represents the upper bound of the Turan density,
    //which we are trying to minimize.
    nVar++;

    //As well as the matrices which we are constraining to be positive semidefinite, we have
    //a series of lower bounds on $\rho$ that must also be satisfied. We encode this in an
    //extra block.
    nBlocks++;

    cout << "Opening the output file......" << flush;
    
    FileOut.open(filenameSDP,ios::trunc);

    if(FileOut.good()==false)
    {
        cout << endl << endl;
        cout << "ERROR: Cannot create the file \"" << filenameSDP << "\"." << endl;
        return false;
    }

    cout << "Done." << endl;
    cout << endl;

    cout << "Creating the SDP problem....." << flush;

    //Output the number of blocks and variables we will be optimizing over.
    FileOut << nVar << endl;
    FileOut << nBlocks << endl;

    //Output the dimensions of the blocks.
    for(i=0;i<noOfTypes;i++)
    for(j=0;j<2;j++)
        if(FInfo[i].BasisSize[j]>0)
            FileOut << FInfo[i].BasisSize[j] << " ";

    //Output the dimension of the block used to constrain $\rho$. We get one constraint for
    //every graph in H[]. The block is diagonal and we encode this information by prefixing a
    //minus sign.
    FileOut << "-" << noOfH << endl;

    //Set the objective function to $\rho$.
    
    //We take the first variable to be $\rho$.
    FileOut << "1";

    //Set the coefficients of all other variables to be 0.
    for(i=0;i<nVar-1;i++)
        FileOut << " 0";

    FileOut << endl;

    //For each variable we have a matrix that consists of blocks down its diagonal. We start
    //by defining such a matrix associated with 0 th variable which is defined as the
    //constant 1. The only constant terms are $d(H_i)$ which appear in the constraints of
    //$\rho$. We multiply the constraints by Factorial[orderOfH] to keep all the coefficients
    //integers. 
    for(i=0;i<noOfH;i++)
        if(NumberOfEdges(H[i])!=0)
        {
            //We only need to set the non-zero entries of the matrix.

            //Set the variable number.
            FileOut << "0 ";

            //Set the block number. Block nBlocks is the block associated with the
            //constraints of $\rho$.
            FileOut << nBlocks << " ";
            
            //Set the position the entry appears in the block. The entries are indexed from 1
            //not 0.
            FileOut << i+1 << " " << i+1 << " ";

            //Set the value of the entry, which is the "density of H[i]" =
            //NumberOfEdges(H[i])/(Factorial[orderOfH] choose 3) multiplied by
            //Factorial[orderOfH] (in order to keep all entries integers).
            FileOut << NumberOfEdges(H[i])*Factorial[orderOfH-3]*6 << endl; 
        }

    //Set the matrix corresponding to variable 1, which we've defined to be $\rho$.
    for(i=0;i<noOfH;i++)
    {
        //Set the variable number.
        FileOut << "1 ";

        //Set the block number. Block nBlocks is the block associated with the constraints of
        //$\rho$.
        FileOut << nBlocks << " ";

        //Set the position the entry appears in the block. The entries are indexed from 1 not
        //0.
        FileOut << i+1 << " " << i+1 << " ";

        //Set the value of the entry, which is 1 multiplied by Factorial[orderOfH], (in order
        //to keep all entries integers).
        FileOut << Factorial[orderOfH] << endl; 
    }

    //Set the matrices for the other variables.

    b = 1; //Current block number.
    v = 2; //Current variable number.
    for(n=0;n<noOfTypes;n++)
    {
        CalculateFlagProducts(FInfo[n]);

        for(m=0;m<2;m++)
        {
            if(FInfo[n].BasisSize[m]<=0)
                continue; //Ignore blocks of dimension 0.

            //r = row index,
            //c = column index.
            for(r=0;r<FInfo[n].BasisSize[m];r++)
            for(c=r;c<FInfo[n].BasisSize[m];c++)
            {
                //Set the position of the variable v, in block b, row r, column c.
                
                //Set the variable and block number.
                FileOut << v << " " << b << " ";

                //Set the position the entry appears in the block.
                FileOut << r+1 << " " << c+1 << " ";

                //Set the value of the entry.
                FileOut << "1" << endl;

                //Multiply basis element r with element c and store the result in product[].

                //Clear product[].
                for(i=0;i<noOfH;i++)
                    product[i] = 0;

                //Multiply flag i of element r with flag j of element c. If m = 1 then we
                //also must take into account the coefficients of the flags which are 1 when
                //i,j = 0, and -1 when i,j = 1. 
                for(i=0;i<FInfo[n].BasisElementSize[m][r];i++)
                for(j=0;j<FInfo[n].BasisElementSize[m][c];j++)
                for(k=0;k<noOfH;k++)
                {
                    if(m==0 || (m==1 && i==j))
                        coeff = 1;
                    else 
                        coeff = -1;

                    product[k] += coeff*pFProd->Data[FInfo[n].BasisElementFlags[m][r][i]]
                                                    [FInfo[n].BasisElementFlags[m][c][j]][k];
                }

                //Variables not on the diagonal make twice the contribution because the
                //matrices are symmetric.
                if(r!=c)
                {
                    for(k=0;k<noOfH;k++)
                        product[k] *= 2;
                }

                //Set the entries in block nBlocks.
                for(k=0;k<noOfH;k++)
                {
                    if(product[k]==0)
                        continue; //We only need to set the non-zero entries.

                    //Set the variable and block number.
                    FileOut << v << " " << nBlocks << " ";

                    //Set the position the entry appears in the block.
                    FileOut << k+1 << " " << k+1 << " ";

                    //Set the value of the entry.
                    FileOut << -product[k] << endl;
                }

                v++;
            }

            b++;
        }
    }

    cout << "Done." << endl;
    cout << endl;

    cout << "Closing the output file......" << flush;

    FileOut.close();

    cout << "Done." << endl;
    cout << endl;

    return true;
}

//Displays Ineq[index]/(Factorial[orderOfH]*4^accuracy) as both a fraction and a decimal.
//This function is used in CalculateDensity() to display an upper bound of the Turan density.
//Returns false if overflow occurs.
bool DisplayDensity(long index)
{
    long      i;
    bool      bNegative;
    bool      bRounded;
    CLargeInt N, D, A, B, C;

    //Set the numerator.
    N = Ineq[index];

    //Calculate the denominator which is equal to Factorial[orderOfH]*4^accuracy.
    LI_Assign(Factorial[orderOfH],D);
    LI_Assign(4,A);
    for(i=0;i<accuracy;i++)
        if(LI_Multiply(A,D,D)==false)
            return false;

    cout << endl;
    cout << " = " << N << " / " << D;

    //Display the Turan density as a decimal to "noOfDecimalPlaces" decimal places.

    //Check if N is negative.
    LI_Assign(0,A);
    if(N.bPositive==false && LI_Compare(A,N)!=0)
        bNegative = true;
    else
        bNegative = false;

    N.bPositive = true;

    //Calculate the digits that will make up the decimal.

    //Calculate 10^noOfDecimalPlaces.
    LI_Assign(10,A);
    LI_Assign(1,C);
    for(i=0;i<noOfDecimalPlaces;i++)
        if(LI_Multiply(A,C,C)==false)
            return false;

    //N *= 10^noOfDecimalPlaces.
    if(LI_Multiply(N,C,N)==false)
        return false;

    //N = N/D.
    LI_Divide(N,D,N,A);

    //N is now the digits that would appear in Ineq[index]/D rounded towards zero. We wish to
    //round up.

    //Check if rounding up is necessary.
    LI_Assign(0,B);
    if(LI_Compare(A,B)!=0)
    {
        //Rounding is necessary, currently the magnitude has been rounded down. If the result
        //is negative then this is fine. If the result is positive then we add one so that
        //the magnitude is rounded up.
        bRounded = true;

        if(bNegative==false)
        {
            LI_Assign(1,A);
            if(LI_Add(A,N,N)==false)
                return false;
        }
    }
    else
        bRounded = false;

    //Start displaying the digits.
    cout << endl;
    cout << " = ";

    if(bNegative==true)
        cout << "-";

    //Display the digits that lie to the left of the decimal point. 
    LI_Divide(N,C,B,N);
    cout << B;
    
    LI_Assign(0,A);
    if(bRounded==false && LI_Compare(A,N)==0)
    {
        cout << endl;
        return true;
    }

    cout << ".";
    
    //Display the digits that lie to the right of the decimal point, one at a time.
    LI_Assign(10,A);
    for(i=0;i<noOfDecimalPlaces;i++)
    {
        //Display the digit (10*N)/C rounded down, where C = 10^noOfDecimalPlaces.
        if(LI_Multiply(N,A,N)==false)
            return false;

        LI_Divide(N,C,B,N);

        //B.Data[0] holds the value of the digit.
        cout << (char)(B.Data[0]+'0');
    }

    if(bRounded==true)
        cout << " (rounded up)";

    cout << endl;

    return true;
}

//Calculates an upper bound of the Turan density and the displays it by calling
//DisplayDensity(). The function takes the values given in Soln[].LMatrix[][] to construct a
//solution to the SDP problem. It calculates the constraints the upper bound must satisfy and
//stores them in Ineq[]. The function is called by main() and expects FInfo[] and H[] to be
//already computed. Returns false if overflow occurs, and true if it was successful.
bool CalculateDensity()
{
    long      i, j, k, m, n;
    long      r, c;
    long      nBlocks;
    long      coeff;
    long      product[maxNoOfGraphs];
    CLargeInt A, B;

    //Calculate an upper bound for the Turan density, based on the solution of the SDP
    //problem which is encoded in Soln[].LMatrix[][].

    cout << "Calculating the bound........" << flush;

    //We are trying to minimize $\rho$ an upper bound of the Turan density. To keep the
    //coefficients in the constraints of $\rho$ integers each inequality is muliplied by
    //Factorial[orderOfH]. The values of the variables in the constraints are the integers
    //given by Soln[].LMatrix^T Soln[].LMatrix divided by 4^accuracy. Hence we multiply the
    //inequality by 4^accuracy to keep the entire expression an integer.
    
    //Calculate 4^accuracy.
    LI_Assign(4,B);
    LI_Assign(1,A);
    
    for(i=0;i<accuracy;i++)
        if(LI_Multiply(B,A,A)==false)
            return false;

    //Multiply A by Factorial[orderOfH]/(orderOfH choose 3).
    LI_Assign(Factorial[orderOfH-3]*6,B);
    
    if(LI_Multiply(A,B,A)==false)
        return false;

    //Initialize Ineq[], with the density of H[] * Factorial[orderOfH] * 4^accuracy, which is
    //equivalent to the number of edges in H[] times A.
    for(i=0;i<noOfH;i++)
    {
        LI_Assign(NumberOfEdges(H[i]),B);
        
        if(LI_Multiply(A,B,Ineq[i])==false)
            return false;
    }

    //Calculate the contribution the entries of Soln[].LMatrix[][] make to Ineq[].
    nBlocks = 0;
    for(n=0;n<noOfTypes;n++)
    {
        CalculateFlagProducts(FInfo[n]);

        for(m=0;m<2;m++)
        {
            if(FInfo[n].BasisSize[m]<=0)
                continue;

            //r = row index,
            //c = column index.
            for(r=0;r<FInfo[n].BasisSize[m];r++)
            for(c=r;c<FInfo[n].BasisSize[m];c++)
            {
                //Multiply basis element r with element c and store the result in product[].

                //Clear product[].
                for(i=0;i<noOfH;i++)
                    product[i] = 0;

                //Multiply flag i of element r with flag j of element c. If m = 1 then we 
                //also must take into account the coefficients of the flags which are 1 when
                //i,j = 0, and -1 when i,j = 1. 
                for(i=0;i<FInfo[n].BasisElementSize[m][r];i++)
                for(j=0;j<FInfo[n].BasisElementSize[m][c];j++)
                for(k=0;k<noOfH;k++)
                {
                    if(m==0 || (m==1 && i==j))
                        coeff = 1;
                    else 
                        coeff = -1;

                    product[k] += coeff*pFProd->Data[FInfo[n].BasisElementFlags[m][r][i]]
                                                    [FInfo[n].BasisElementFlags[m][c][j]][k];
                }

                //Variables not on the diagonal make twice the contribution because the
                //matrices are symmetric. 
                if(r!=c)
                {
                    for(k=0;k<noOfH;k++)
                        product[k] *= 2;
                }

                //Calculate the value of the r, c entry of 
                //Soln[nBlocks].LMatrix^T * Soln[nBlocks].LMatrix and store it in A.
                LI_Assign(0,A);
                for(k=0;k<FInfo[n].BasisSize[m];k++)
                {
                    if(LI_Multiply(Soln[nBlocks].LMatrix[k][r],
                                   Soln[nBlocks].LMatrix[k][c],B)==false)
                        return false;

                    if(LI_Add(B,A,A)==false)
                        return false;
                }

                //Add A*product[] to Ineq[].
                for(k=0;k<noOfH;k++)
                {
                    LI_Assign(product[k],B);

                    if(LI_Multiply(B,A,B)==false)
                        return false;

                    if(LI_Add(B,Ineq[k],Ineq[k])==false)
                        return false;
                }
            }

            nBlocks++;
        }
    }

    //Determine which is the largest lower bound for $\rho$. Store the index in j.
    j = 0;
    for(i=0;i<noOfH;i++)
        if(LI_Compare(Ineq[i],Ineq[j])==1)
            j = i;

    cout << "Done." << endl;
    cout << endl;

    //Display the smallest upper bound for the Turan density.
    cout << "Upper bound for the Turan density";

    if(DisplayDensity(j)==false)
        return false;

    cout << endl;

    return true;
}

int main()
{
    long i, j, k;
    long e, n;
    long choice, choice2;

    //Get user input.
    choice  = 0;
    choice2 = 0;

    cout << "1: Create a \".dat-s\" file as input to a SDP solver." << endl;
    cout << endl;
    cout << "2: Rationalize the output of the solver and store" << endl;
    cout << "   the result in a \".soln\" file." << endl;
    cout << endl;
    cout << "3: Calculate an upper bound of the Turan density" << endl;
    cout << "   using a \".soln\" file." << endl;
    cout << endl;
    cout << endl;
    cout << "\? ";

    cin >> choice;

    cout << endl;

    if((choice!=1 && choice!=2 && choice!=3) || cin.good()==false) 
    {
        cout << "ERROR: Unexpected input." << endl;
        cout << endl;

        return 0;
    }

    //Allocate memory, and initialize Factorial[] and Iso[][].
    if(Init()==false)
    {
        cout << "ERROR: Init() failed. Ran out of memory." << endl;
        cout << endl;

        Cleanup();
        return 0;
    }

    //Initialize ForbiddenGraph[], from ForbiddenOrder[], and ForbiddenEdgeSet[][].
    for(n=0;n<noOfForbiddenGraphs;n++)
    {
        //Set the number of vertices.
        ForbiddenGraph[n].nV = ForbiddenOrder[n];

        //Clear the adjacency matrix.
        for(i=0;i<orderOfH;i++)
        for(j=0;j<orderOfH;j++)
        for(k=0;k<orderOfH;k++)
            ForbiddenGraph[n].M[i][j][k] = false;

        //The edge ForbiddenEdgeSet[n][e] is stored as an integer whose digits in base 10
        //indicate the vertices. The end of the edge set is indicated by the value 0.
        e = 0;
        while(ForbiddenEdgeSet[n][e]!=0)
        {
            //Because C++ indexes arrays from 0 it will be simpler to take our vertex set to
            //be 0, 1, ..., ForbiddenOrder[n]-1, rather than 1, 2, ..., ForbiddenOrder[n].
            //Hence the edge 124 becomes the edge {0,1,3}. In general
            //ForbiddenEdgeSet[n][e] = (k+1)*100+(j+1)*10+(i+1).
            i = ForbiddenEdgeSet[n][e]%10-1;
            j = (ForbiddenEdgeSet[n][e]/10)%10-1;
            k = ForbiddenEdgeSet[n][e]/100-1;

            if(i>=orderOfH || j>=orderOfH || k>=orderOfH)
            {
                cout << "ERROR: ForbiddenEdgeSet[][] is incorrectly formatted." << endl;
                cout << "Try increasing orderOfH." << endl;
                cout << endl;

                Cleanup();
                return 0;
            }

            //Update the adjacency matrix.
            ForbiddenGraph[n].M[i][j][k] = true;
            ForbiddenGraph[n].M[i][k][j] = true;
            ForbiddenGraph[n].M[j][i][k] = true;
            ForbiddenGraph[n].M[j][k][i] = true;
            ForbiddenGraph[n].M[k][i][j] = true;
            ForbiddenGraph[n].M[k][j][i] = true;

            e++;
        }
    }

    //Initialize FInfo[].Type, from TypeOrder[], and TypeEdgeSet[][].
    for(n=0;n<noOfTypes;n++)
    {
        //Set the number of vertices.
        FInfo[n].Type.nV = TypeOrder[n];

        //Clear the adjacency matrix.
        for(i=0;i<orderOfH;i++)
        for(j=0;j<orderOfH;j++)
        for(k=0;k<orderOfH;k++)
            FInfo[n].Type.M[i][j][k] = false;

        //Update the adjacency matrix as we did for ForbiddenGraph[].
        e = 0;
        while(TypeEdgeSet[n][e]!=0)
        {
            i = TypeEdgeSet[n][e]%10-1;
            j = (TypeEdgeSet[n][e]/10)%10-1;
            k = TypeEdgeSet[n][e]/100-1;
            
            if(i>=orderOfH || j>=orderOfH || k>=orderOfH)
            {
                cout << "ERROR: TypeEdgeSet[][] is incorrectly formatted." << endl;
                cout << "Try increasing orderOfH." << endl;
                cout << endl;

                Cleanup();
                return 0;
            }

            FInfo[n].Type.M[i][j][k] = true;
            FInfo[n].Type.M[i][k][j] = true;
            FInfo[n].Type.M[j][i][k] = true;
            FInfo[n].Type.M[j][k][i] = true;
            FInfo[n].Type.M[k][i][j] = true;
            FInfo[n].Type.M[k][j][i] = true;

            e++;
        }
    }

    //Check the Types are admissible.
    for(i=0;i<noOfTypes;i++)
    for(j=0;j<noOfForbiddenGraphs;j++)
        if(IsSubgraph(ForbiddenGraph[j],FInfo[i].Type,bForbidOnlyInducedSubgraphs)==true)
        {
            cout << "ERROR: The type (" << i << ")" << endl;
            cout << FInfo[i].Type << endl;
            cout << "contains the forbidden graph (" << j << ")" << endl;
            cout << ForbiddenGraph[j] << endl;
            
            if(bForbidOnlyInducedSubgraphs==true)
                cout << "as an induced subgraph." << endl;
            else
                cout << "as a subgraph." << endl;

            cout << endl;
            
            Cleanup();
            return 0;
        }

    //Initialize FInfo[].
    cout << "Generating flag data........." << flush;
    
    for(i=0;i<noOfTypes;i++)
    {
        //Check FlagOrder[i] isn't too big.
        if(FlagOrder[i]>(orderOfH+FInfo[i].Type.nV)/2)
        {
            cout << endl;
            cout << endl;
            cout << "ERROR: FlagOrder[" << i << "] is too large." << endl;
            cout << "Try increasing orderOfH or decreasing FlagOrder[" << i << "]." << endl;
            cout << endl;
            
            Cleanup();
            return 0;
        }

        //Generate FInfo[i].FlagList[].
        if(GenerateFlags(FInfo[i],FlagOrder[i])==false)
        {
            cout << endl;
            cout << endl;
            cout << "ERROR: ";

            if(maxNoOfFlags<=maxNoOfGraphs)
                cout << "Try increasing maxNoOfFlags." << endl;
            else
                cout << "Try increasing maxNoOfGraphs." << endl;

            cout << endl;

            Cleanup();
            return 0;
        }

        //Generate a new basis.
        GenerateBasis(FInfo[i]);
    }

    cout << "Done." << endl;

    for(i=0;i<noOfTypes;i++)
    {
        cout << "Generated " << setw(3) << FInfo[i].noOfFlags;
        
        if(FInfo[i].noOfFlags!=1)
            cout << " Type[" << i << "]-flags." << endl;
        else
            cout << " Type[" << i << "]-flag." << endl;

    }

    cout << endl;

    if(choice==2)
    {
        //Create the ".soln" file.
        if(CreateSolnFile()==false)
        {
            cout << endl;
            Cleanup();
            return 0;
        }

        //Get user input.
        cout << endl;
        cout << "1: Calculate an upper bound of the Turan density." << endl;
        cout << endl;
        cout << "2: Exit" << endl;
        cout << endl;
        cout << endl;
        cout << "\? ";

        cin >> choice2;

        cout << endl;

        if((choice2!=1 && choice2!=2) || cin.good()==false) 
        {
            cout << "ERROR: Unexpected input." << endl;
            cout << endl;

            Cleanup();
            return 0;
        }
    }

    if(choice==1 || choice==3 || (choice==2 && choice2==1))
    {
        //Initialize H[].
        cout << "Generating H[]..............." << flush;
        if(GenerateH()==false)
        {
            cout << endl;
            cout << endl;
            cout << "ERROR: Try increasing maxNoOfGraphs." << endl;
            cout << endl;

            Cleanup();
            return 0;
        }

        cout << "Done." << endl;
        cout << "Generated " << noOfH << " graphs." << endl;
        cout << endl;
    }
    
    if(choice==1)
    {
        //Create the ".dat-s" file that is to be used as input to a SDP solver.
        if(CreateSDPFile()==false)
        {
            cout << endl;
            Cleanup();
            return 0;
        }
    }

    if(choice==3)
    {
        //Initialize Soln[].LMatrix[][] with the precomputed values stored in a ".soln" file.
        if(ReadSolnFile()==false)
        {
            cout << endl;
            Cleanup();
            return 0;
        }
    }

    if(choice==3 || (choice==2 && choice2==1))
    {
        //Calculate an upper bound for the Turan density using the values in
        //Soln[].LMatrix[][]
        if(CalculateDensity()==false)
        {
            //Overflow occured.
            cout << endl << endl;
            cout << "ERROR: Try changing LI_MAX." << endl;
            cout << endl;

            Cleanup();
            return 0;
        }
    }

    cout << endl;
    cout << "------------ Finished ------------" << endl;
    cout << endl;

    Cleanup();
    return 0;
}
